import numpy as np
from numpy import newaxis as na
from scipy.spatial import ConvexHull
from topfarm.constraint_components import Constraint, ConstraintComponent
from topfarm.utils import smooth_max, smooth_max_gradient
import topfarm
from shapely.geometry import Polygon, MultiPolygon, LineString
from shapely.ops import unary_union
import warnings


class XYBoundaryConstraint(Constraint):
    def __init__(self, boundary, boundary_type='convex_hull', units=None, relaxation=False, **kwargs):
        """Initialize XYBoundaryConstraint

        Parameters
        ----------
        boundary : array_like (n,2) or list of tuples (array_like (n,2), boolean)
            boundary coordinates. If boundary is array_like (n,2) it indicates a single boundary and can be used with
            boundary types: 'convex_hull', 'polygon', 'rectangle','square'. If boundary is list of tuples (array_like (n,2), boolean),
            it is multiple boundaries where the boolean is 1 for inclusion zones and 0 for exclusion zones and can be used with the
            boundary type: 'multi_polygon'.
        boundary_type : 'convex_hull', 'polygon', 'rectangle','square'
            - 'convex_hull' (default): Convex hul around boundary points\n
            - 'polygon': Polygon boundary (may be non convex). Less suitable for gradient-based optimization\n
            - 'rectangle': Smallest axis-aligned rectangle covering the boundary points\n
            - 'square': Smallest axis-aligned square covering the boundary points
            - 'multi_polygon': Mulitple polygon boundaries incl. exclusion zones (may be non convex).\n
            - 'turbine_specific': Set of multiple polygon boundaries that depend on the wind turbine type. \n


        """
        if boundary_type == 'multi_polygon':
            self.zones = boundary
            self.boundary = np.asarray(self.zones[0].boundary)
        elif boundary_type == 'turbine_specific':
            self.zones = boundary
            assert 'turbines' in list(kwargs)
            self.turbines = kwargs['turbines']
            self.boundary = np.asarray(self.zones[0].boundary)
        else:
            self.boundary = np.asarray(boundary)
        self.boundary_type = boundary_type
        self.const_id = 'xyboundary_comp_{}_{}'.format(boundary_type, int(self.boundary.sum()))
        self.units = units
        self.relaxation = relaxation

    def get_comp(self, n_wt):
        if not hasattr(self, 'boundary_comp'):
            if self.boundary_type == 'polygon':
                self.boundary_comp = PolygonBoundaryComp(
                    n_wt, self.boundary, self.const_id, self.units, self.relaxation)
            elif self.boundary_type == 'multi_polygon':
                self.boundary_comp = MultiPolygonBoundaryComp(n_wt, self.zones, const_id=self.const_id, units=self.units, relaxation=self.relaxation)
            elif self.boundary_type == 'turbine_specific':
                self.boundary_comp = TurbineSpecificBoundaryComp(n_wt, self.turbines, self.zones, const_id=self.const_id, units=self.units, relaxation=self.relaxation)
            else:
                self.boundary_comp = ConvexBoundaryComp(n_wt, self.boundary, self.boundary_type, self.const_id, self.units)
        return self.boundary_comp

    @property
    def constraintComponent(self):
        return self.boundary_comp

    def set_design_var_limits(self, design_vars):
        if self.boundary_type in ['multi_polygon', 'turbine_specific']:
            bound_min = np.vstack([(bound).min(0) for bound, _ in self.boundary_comp.boundaries]).min(0)
            bound_max = np.vstack([(bound).max(0) for bound, _ in self.boundary_comp.boundaries]).max(0)
        else:
            bound_min = self.boundary_comp.xy_boundary.min(0)
            bound_max = self.boundary_comp.xy_boundary.max(0)
        for k, l, u in zip([topfarm.x_key, topfarm.y_key], bound_min, bound_max):
            if k in design_vars:
                if len(design_vars[k]) == 4:
                    design_vars[k] = (design_vars[k][0], np.maximum(design_vars[k][1], l),
                                      np.minimum(design_vars[k][2], u), design_vars[k][-1])
                else:
                    design_vars[k] = (design_vars[k][0], l, u, design_vars[k][-1])

    def _setup(self, problem, group='pre_constraints'):
        n_wt = problem.n_wt
        self.boundary_comp = self.get_comp(n_wt)
        self.boundary_comp.problem = problem
        self.set_design_var_limits(problem.design_vars)
        # problem.xy_boundary = np.r_[self.boundary_comp.xy_boundary, self.boundary_comp.xy_boundary[:1]]
        problem.indeps.add_output('xy_boundary', self.boundary_comp.xy_boundary)
        getattr(problem.model, group).add_subsystem('xy_bound_comp', self.boundary_comp, promotes=['*'])

    def setup_as_constraint(self, problem, group='pre_constraints'):
        self._setup(problem, group=group)
        problem.model.add_constraint('boundaryDistances', lower=self.boundary_comp.zeros)

    def setup_as_penalty(self, problem, group='pre_constraints'):
        self._setup(problem, group=group)


class CircleBoundaryConstraint(XYBoundaryConstraint):
    def __init__(self, center, radius):
        """Initialize CircleBoundaryConstraint

        Parameters
        ----------
        center : (float, float)
            center position (x,y)
        radius : int or float
            circle radius
        """

        self.center = np.array(center)
        self.radius = radius
        self.const_id = 'circle_boundary_comp_{}_{}'.format(
            '_'.join([str(int(c)) for c in center]), int(radius)).replace('.', '_')

    def get_comp(self, n_wt):
        if not hasattr(self, 'boundary_comp'):
            self.boundary_comp = CircleBoundaryComp(n_wt, self.center, self.radius, self.const_id)
        return self.boundary_comp

    def set_design_var_limits(self, design_vars):
        for k, l, u in zip([topfarm.x_key, topfarm.y_key],
                           self.center - self.radius,
                           self.center + self.radius):
            if len(design_vars[k]) == 4:
                design_vars[k] = (design_vars[k][0], np.maximum(design_vars[k][1], l),
                                  np.minimum(design_vars[k][2], u), design_vars[k][-1])
            else:
                design_vars[k] = (design_vars[k][0], l, u, design_vars[k][-1])


class BoundaryBaseComp(ConstraintComponent):
    def __init__(self, n_wt, xy_boundary=None, const_id=None, units=None, relaxation=False, **kwargs):
        super().__init__(**kwargs)
        self.n_wt = n_wt
        self.xy_boundary = np.array(xy_boundary)
        self.const_id = const_id
        self.units = units
        self.relaxation = relaxation
        if np.any(self.xy_boundary[0] != self.xy_boundary[-1]):
            self.xy_boundary = np.r_[self.xy_boundary, self.xy_boundary[:1]]

    def setup(self):
        # Explicitly size input arrays
        self.add_input(topfarm.x_key, np.zeros(self.n_wt),
                       desc='x coordinates of turbines in global ref. frame', units=self.units)
        self.add_input(topfarm.y_key, np.zeros(self.n_wt),
                       desc='y coordinates of turbines in global ref. frame', units=self.units)
        if self.relaxation:
            self.add_input('time', 0)
        if hasattr(self, 'types'):
            self.add_input('type', np.zeros(self.n_wt))
        self.add_output('penalty_' + self.const_id, val=0.0)
        # Explicitly size output array
        # (vector with positive elements if turbines outside of hull)
        self.add_output('boundaryDistances', self.zeros,
                        desc="signed perpendicular distances from each turbine to each face CCW; + is inside")
        self.declare_partials('boundaryDistances', [topfarm.x_key, topfarm.y_key])
        if self.relaxation:
            self.declare_partials('boundaryDistances', 'time')

        # self.declare_partials('boundaryDistances', ['boundaryVertices', 'boundaryNormals'], method='fd')

    def compute(self, inputs, outputs):
        # calculate distances from each point to each face
        args = {x: inputs[x] for x in [topfarm.x_key, topfarm.y_key, topfarm.type_key] if x in inputs}
        boundaryDistances = self.distances(**args)
        outputs['boundaryDistances'] = boundaryDistances
        outputs['penalty_' + self.const_id] = np.sum(np.minimum(boundaryDistances, 0) ** 2)

    def compute_partials(self, inputs, partials):
        # return Jacobian dict
        if not self.relaxation:
            dx, dy = self.gradients(**{xy: inputs[k] for xy, k in zip('xy', [topfarm.x_key, topfarm.y_key])})
        else:
            dx, dy, dt = self.gradients(**{xy: inputs[k] for xy, k in zip('xy', [topfarm.x_key, topfarm.y_key])})

        partials['boundaryDistances', topfarm.x_key] = dx
        partials['boundaryDistances', topfarm.y_key] = dy
        if self.relaxation:
            partials['boundaryDistances', 'time'] = dt

    def plot(self, ax):
        """Plot boundary"""
        if isinstance(self, TurbineSpecificBoundaryComp):
            linestyles = ['--', '-']
            colors = np.array(['b', 'r', 'm', 'c', 'g', 'y', 'orange', 'indigo', 'grey'] * 10)
            for n, t in enumerate(self.types):
                line, = ax.plot(*self.ts_merged_xy_boundaries[n][0][0][0, :], color=colors[t], linewidth=1, label=f'{self.wind_turbines._names[n]} boundary')
                for bound, io in self.ts_merged_xy_boundaries[n]:
                    ax.plot(np.asarray(bound)[:, 0].tolist() + [np.asarray(bound)[0, 0]],
                            np.asarray(bound)[:, 1].tolist() + [np.asarray(bound)[0, 1]], color=colors[t], linewidth=1, linestyle=linestyles[io])
        elif isinstance(self, MultiPolygonBoundaryComp):
            colors = ['--k', 'k']
            if self.relaxation != 0:
                for bound, io in self.relaxed_polygons():
                    ax.plot(np.asarray(bound)[:, 0].tolist() + [np.asarray(bound)[0, 0]],
                            np.asarray(bound)[:, 1].tolist() + [np.asarray(bound)[0, 1]], c='r', linewidth=1, linestyle='--')
                ax.plot([], c='r', linewidth=1, linestyle='--', label='Relaxed boundaries')
            for bound, io in self.boundaries:
                ax.plot(np.asarray(bound)[:, 0].tolist() + [np.asarray(bound)[0, 0]],
                        np.asarray(bound)[:, 1].tolist() + [np.asarray(bound)[0, 1]], colors[io], linewidth=1)
        else:
            ax.plot(self.xy_boundary[:, 0].tolist() + [self.xy_boundary[0, 0]],
                    self.xy_boundary[:, 1].tolist() + [self.xy_boundary[0, 1]], 'k', linewidth=1)


class ConvexBoundaryComp(BoundaryBaseComp):
    def __init__(self, n_wt, xy_boundary=None, boundary_type='convex_hull', const_id=None, units=None):
        self.boundary_type = boundary_type
#        self.const_id = const_id
        self.calculate_boundary_and_normals(xy_boundary)
        super().__init__(n_wt, self.xy_boundary, const_id, units)
        self.calculate_gradients()
        self.zeros = np.zeros([self.n_wt, self.nVertices])
#        self.units = units

    def calculate_boundary_and_normals(self, xy_boundary):
        xy_boundary = np.asarray(xy_boundary)
        if self.boundary_type == 'convex_hull':
            # find the points that actually comprise a convex hull
            hull = ConvexHull(list(xy_boundary))

            # keep only xy_vertices that actually comprise a convex hull and arrange in CCW order
            self.xy_boundary = xy_boundary[hull.vertices]
        elif self.boundary_type == 'square':
            min_ = xy_boundary.min(0)
            max_ = xy_boundary.max(0)
            range_ = (max_ - min_)
            x_c, y_c = min_ + range_ / 2
            r = range_.max() / 2
            self.xy_boundary = np.array([(x_c - r, y_c - r), (x_c + r, y_c - r),
                                         (x_c + r, y_c + r), (x_c - r, y_c + r)])
        elif self.boundary_type == 'rectangle':
            min_ = xy_boundary.min(0)
            max_ = xy_boundary.max(0)
            range_ = (max_ - min_)
            x_c, y_c = min_ + range_ / 2
            r = range_ / 2
            self.xy_boundary = np.array([(x_c - r[0], y_c - r[1]), (x_c + r[0], y_c - r[1]),
                                         (x_c + r[0], y_c + r[1]), (x_c - r[0], y_c + r[1])])
        else:
            raise NotImplementedError("Boundary type '%s' is not implemented" % self.boundary_type)

        # get the real number of xy_vertices
        self.nVertices = self.xy_boundary.shape[0]

        # initialize normals array
        unit_normals = np.zeros([self.nVertices, 2])

        # determine if point is inside or outside of each face, and distances from each face
        for j in range(0, self.nVertices):

            # calculate the unit normal vector of the current face (taking points CCW)
            if j < self.nVertices - 1:  # all but the set of point that close the shape
                normal = np.array([self.xy_boundary[j + 1, 1] - self.xy_boundary[j, 1],
                                   -(self.xy_boundary[j + 1, 0] - self.xy_boundary[j, 0])])
                unit_normals[j] = normal / np.linalg.norm(normal)
            else:   # the set of points that close the shape
                normal = np.array([self.xy_boundary[0, 1] - self.xy_boundary[j, 1],
                                   -(self.xy_boundary[0, 0] - self.xy_boundary[j, 0])])
                unit_normals[j] = normal / np.linalg.norm(normal)

        self.unit_normals = unit_normals

    def calculate_gradients(self):
        unit_normals = self.unit_normals

        # initialize array to hold distances from each point to each face
        dfaceDistance_dx = np.zeros([self.n_wt * self.nVertices, self.n_wt])
        dfaceDistance_dy = np.zeros([self.n_wt * self.nVertices, self.n_wt])

        for i in range(0, self.n_wt):
            # determine if point is inside or outside of each face, and distances from each face
            for j in range(0, self.nVertices):

                # define the derivative vectors from the point of interest to the first point of the face
                dpa_dx = np.array([-1.0, 0.0])
                dpa_dy = np.array([0.0, -1.0])

                # find perpendicular distances derivatives from point to current surface (vector projection)
                ddistanceVec_dx = np.vdot(dpa_dx, unit_normals[j]) * unit_normals[j]
                ddistanceVec_dy = np.vdot(dpa_dy, unit_normals[j]) * unit_normals[j]

                # calculate derivatives for the sign of perpendicular distances from point to current face
                dfaceDistance_dx[i * self.nVertices + j, i] = np.vdot(ddistanceVec_dx, unit_normals[j])
                dfaceDistance_dy[i * self.nVertices + j, i] = np.vdot(ddistanceVec_dy, unit_normals[j])

        # return Jacobian dict
        self.dfaceDistance_dx = dfaceDistance_dx
        self.dfaceDistance_dy = dfaceDistance_dy

    def calculate_distance_to_boundary(self, points):
        """
        :param points: points that you want to calculate the distances from to the faces of the convex hull
        :return face_distace: signed perpendicular distances from each point to each face; + is inside
        """

        nPoints = np.array(points).shape[0]
        xy_boundary = self.xy_boundary[:-1]
        nVertices = xy_boundary.shape[0]
        vertices = xy_boundary
        unit_normals = self.unit_normals
        # initialize array to hold distances from each point to each face
        face_distance = np.zeros([nPoints, nVertices])
        from numpy import newaxis as na

        # define the vector from the point of interest to the first point of the face
        PA = (vertices[:, na] - points[na])

        # find perpendicular distances from point to current surface (vector projection)
        dist = np.sum(PA * unit_normals[:, na], 2)
        # calculate the sign of perpendicular distances from point to current face (+ is inside, - is outside)
        d_vec = dist[:, :, na] * unit_normals[:, na]
        face_distance = np.sum(d_vec * unit_normals[:, na], 2)
        return face_distance.T

    def distances(self, x, y):
        return self.calculate_distance_to_boundary(np.array([x, y]).T)

    def gradients(self, x, y):
        return self.dfaceDistance_dx, self.dfaceDistance_dy

    def satisfy(self, state, pad=1.1):
        x, y = [np.asarray(state[xyz], dtype=float) for xyz in [topfarm.x_key, topfarm.y_key]]
        dist = self.distances(x, y)
        dist = np.where(dist < 0, np.minimum(dist, -.01), dist)
        dx, dy = self.gradients(x, y)  # independent of position
        dx = dx[:self.nVertices, 0]
        dy = dy[:self.nVertices, 0]
        for i in np.where(dist.min(1) < 0)[0]:  # loop over turbines that violate edges
            # find smallest movement that where the constraints are satisfied
            d = dist[i]
            v = np.linspace(-np.abs(d.min()), np.abs(d.min()), 100)
            X, Y = np.meshgrid(v, v)
            m = np.ones_like(X)
            for dx_, dy_, d in zip(dx, dy, dist.T):
                m = np.logical_and(m, X * dx_ + Y * dy_ >= -d[i])
            index = np.argmin(X[m]**2 + Y[m]**2)
            x[i] += X[m][index]
            y[i] += Y[m][index]
        state[topfarm.x_key] = x
        state[topfarm.y_key] = y
        return state


class PolygonBoundaryComp(BoundaryBaseComp):
    def __init__(self, n_wt, xy_boundary, const_id=None, units=None, relaxation=False):

        self.nTurbines = n_wt
        self.const_id = const_id
        self.zeros = np.zeros(self.nTurbines)
        self.units = units
        self.boundary_properties = self.get_boundary_properties(xy_boundary)
        BoundaryBaseComp.__init__(self, n_wt, xy_boundary=self.boundary_properties[0], const_id=const_id,
                                  units=units, relaxation=relaxation)
        self._cache_input = None
        self._cache_output = None
        self.relaxation = relaxation

    def get_boundary_properties(self, xy_boundary, inclusion_zone=True):
        vertices = np.array(xy_boundary)

        def get_edges(vertices, counter_clockwise):
            if np.any(vertices[0] != vertices[-1]):
                vertices = np.r_[vertices, vertices[:1]]
            x1, y1 = A = vertices[:-1].T
            x2, y2 = B = vertices[1:].T
            double_area = np.sum((x1 - x2) * (y1 + y2))  # 2 x Area (+: counterclockwise
            assert double_area != 0, "Area must be non-zero"
            if (counter_clockwise and double_area < 0) or (not counter_clockwise and double_area > 0):  #
                return get_edges(vertices[::-1], counter_clockwise)
            else:
                return vertices[:-1], A, B

        # inclusion zones are defined counter clockwise (unit-normal vector pointing in) while
        # exclusion zones are defined clockwise (unit-normal vector pointing out)
        xy_boundary, A, B = get_edges(vertices, inclusion_zone)

        dx, dy = AB = B - A
        AB_len = np.linalg.norm(AB, axis=0)
        edge_unit_normal = (np.array([-dy, dx]) / AB_len)

        # A_normal and B_normal are the normal vectors at the nodes A,B (the mean of the adjacent edge normal vectors
        A_normal = (edge_unit_normal + np.roll(edge_unit_normal, 1, 1)) / 2
        B_normal = np.roll(A_normal, -1, 1)

        # import matplotlib.pyplot as plt
        # for (x, y), (dx, dy), (unx, uny) in zip(A.T, AB.T, edge_unit_normal.T):
        #     plt.arrow(x, y, dx, dy, color='k', head_width=.2)
        #     plt.arrow(x, y, unx, uny, color='r', head_width=.2)
        # for (x, y), (nx, ny) in zip(A.T, A_normal.T):
        #     plt.arrow(x, y, nx, ny, color='b', head_width=.2)
        # for (x, y), (nx, ny) in zip(B.T, B_normal.T):
        #     plt.arrow(x, y, nx / 2, ny / 2, color='g', head_width=.2)

        return (xy_boundary, A, B, AB, AB_len, edge_unit_normal, A_normal, B_normal)

    def _calc_distance_and_gradients(self, x, y, boundary_properties=None):
        """
        distances point, P=(x,y) to edge(A->B)
        +/-: inside/outside
        """
        def vec_len(vec):
            return np.linalg.norm(vec, axis=0)

        boundary_properties = boundary_properties or self.boundary_properties[1:]
        A, B, AB, AB_len, edge_unit_normal, A_normal, B_normal = boundary_properties
        """
        A: edge start point
        B: edge end point
        edge_unit_normal: unit vector perpendicular to edge pointing to the good side
        (i.e. inside for inclusion zones and outside for exclusion zones)
        AB: Vector from A to B (edge)
        AB_len: length of AB (edge)
        A_normal: mean of edge unit normal vectors adjacent to A
        B_normal: mean of edge unit normal vectors adjacent to B
        """

        # Add dim to match (2, #P, #Edges), where the first dimension is (x,y)
        P = np.array([x, y])[:, :, na]
        A, B, AB = A[:, na], B[:, na], AB[:, na]
        edge_unit_normal, A_normal, B_normal = edge_unit_normal[:, na], A_normal[:, na], B_normal[:, na]
        AB_len = AB_len[na]

        # ===============================================================================================================
        # Determine if P is closer to A, B or the edge (between A and B)
        # ===============================================================================================================
        AP = P - A  # vector from edge start to point
        BP = P - B  # vector from edge end to point

        # signed component of AP on the edge vector
        a_tilde = np.sum(AP * AB, axis=0) / AB_len

        # a_tilde < 0: closer to A
        # a_tilde > |AB|: closer to B
        # else: closer to edge (between A and B)
        use_A = 0 > a_tilde
        use_B = a_tilde > AB_len

        # ===============================================================================================================
        # Calculate distance from P to closer point on edge
        # ===============================================================================================================

        # Perpendicular distances to edge (AP dot edge_unit_normal product).
        # This is the distance to the edge if not use_A or use_B
        distance = np.sum((AP) * edge_unit_normal, 0)

        # Update distance for points closer to A
        good_side_of_A = (np.sum((AP * A_normal)[:, use_A], 0) > 0)
        sign_use_A = np.where(good_side_of_A, 1, -1)
        distance[use_A] = (vec_len(AP[:, use_A]) * sign_use_A)

        # Update distance for points closer to B
        good_side_of_B = np.sum((BP * B_normal)[:, use_B], 0) > 0
        sign_use_B = np.where(good_side_of_B, 1, -1)
        distance[use_B] = (vec_len(BP[:, use_B]) * sign_use_B)

        # ===============================================================================================================
        # Calculate gradient of distance from P to closer point on edge wrt. x and y
        # ===============================================================================================================

        # Gradient of perpendicular distances to edge.
        # This is the gradient if not use_A or use_B
        ddist_dxy = np.tile(edge_unit_normal, (1, len(x), 1))

        # Update gradient for points closer to A or B
        ddist_dxy[:, use_A] = sign_use_A * (AP[:, use_A] / vec_len(AP[:, use_A]))
        ddist_dxy[:, use_B] = sign_use_B * (BP[:, use_B] / vec_len(BP[:, use_B]))
        ddist_dX, ddist_dY = ddist_dxy

        return distance, ddist_dX, ddist_dY

    def calc_distance_and_gradients(self, x, y):
        if np.all(np.array([x, y]) == self._cache_input):
            return self._cache_output
        distance, ddist_dX, ddist_dY = self._calc_distance_and_gradients(x, y)
        closest_edge_index = np.argmin(np.abs(distance), 1)
        self._cache_input = np.array([x, y])
        self._cache_output = [np.choose(closest_edge_index, v.T) for v in [distance, ddist_dX, ddist_dY]]
        return self._cache_output

    def distances(self, x, y):
        return self.calc_distance_and_gradients(x, y)[0]

    def gradients(self, x, y):
        _, dx, dy = self.calc_distance_and_gradients(x, y)
        return np.diagflat(dx), np.diagflat(dy)

    def satisfy(self, state, pad=1.1):
        x, y = [np.asarray(state[xy], dtype=float) for xy in [topfarm.x_key, topfarm.y_key]]
        dist = self.distances(x, y)
        dx, dy = map(np.diag, self.gradients(x, y))
        m = dist < 0
        x[m] -= dx[m] * dist[m] * pad
        y[m] -= dy[m] * dist[m] * pad
        state[topfarm.x_key] = x
        state[topfarm.y_key] = y
        return state


class CircleBoundaryComp(PolygonBoundaryComp):
    def __init__(self, n_wt, center, radius, const_id=None, units=None):
        self.center = center
        self.radius = radius
        t = np.linspace(0, 2 * np.pi, 100)
        xy_boundary = self.center + np.array([np.cos(t), np.sin(t)]).T * self.radius
        BoundaryBaseComp.__init__(self, n_wt, xy_boundary, const_id, units)
        self.zeros = np.zeros(self.n_wt)

    def plot(self, ax=None):
        from matplotlib.pyplot import Circle
        import matplotlib.pyplot as plt
        ax = ax or plt.gca()
        circle = Circle(self.center, self.radius, color='k', fill=False)
        ax.add_artist(circle)

    def distances(self, x, y):
        return self.radius - np.sqrt((x - self.center[0])**2 + (y - self.center[1])**2)

    def gradients(self, x, y):
        theta = np.arctan2(y - self.center[1], x - self.center[0])
        dx = -1 * np.ones_like(x)
        dy = -1 * np.ones_like(x)
        dist = self.radius - np.sqrt((x - self.center[0])**2 + (y - self.center[1])**2)
        not_center = dist != self.radius
        dx[not_center], dy[not_center] = -np.cos(theta[not_center]), -np.sin(theta[not_center])
        return np.diagflat(dx), np.diagflat(dy)


class Zone(object):
    def __init__(self, boundary, dist2wt, geometry_type, incl, name):
        self.name = name
        self.boundary = boundary
        self.dist2wt = dist2wt
        self.geometry_type = geometry_type
        self.incl = incl


class InclusionZone(Zone):
    def __init__(self, boundary, dist2wt=None, geometry_type='polygon', name=''):
        super().__init__(np.asarray(boundary), dist2wt, geometry_type, incl=1, name=name)


class ExclusionZone(Zone):
    def __init__(self, boundary, dist2wt=None, geometry_type='polygon', name=''):
        super().__init__(np.asarray(boundary), dist2wt, geometry_type, incl=0, name=name)


class MultiPolygonBoundaryComp(PolygonBoundaryComp):
    def __init__(self, n_wt, zones, const_id=None, units=None, relaxation=False, method='nearest',
                 simplify_geometry=False):
        '''
        Parameters
        ----------
        n_wt : TYPE
            DESCRIPTION.
        zones : list
            list of InclusionZone and ExclusionZone objects
        const_id : TYPE, optional
            DESCRIPTION. The default is None.
        units : TYPE, optional
            DESCRIPTION. The default is None.
        method : {'nearest' or 'smooth_min'}, optional
            'nearest' calculate the distance to the nearest edge or point'smooth_min'
            calculates the weighted minimum distance to all edges/points. The default is 'nearest'.
        simplify : float or dict
            if float, simplification tolerance. if dict, shapely.simplify keyword arguments
        Returns
        -------
        None.

        '''
        self.zones = zones
        self.bounds_poly, xy_boundaries = self.get_xy_boundaries()
        PolygonBoundaryComp.__init__(self, n_wt, xy_boundary=xy_boundaries[0], const_id=const_id, units=units, relaxation=relaxation)
        # self.bounds_poly = [Polygon(x) for x in xy_boundaries]
        self.incl_excls = [x.incl for x in zones]
        self._setup_boundaries()
        self.relaxation = relaxation
        self.method = method
        if simplify_geometry:
            self.simplify(simplify_geometry)

    def simplify(self, simplify_geometry):
        if isinstance(simplify_geometry, dict):
            self.bounds_poly = [rp.simplify(**simplify_geometry) for rp in self.bounds_poly]
        else:
            self.bounds_poly = [rp.simplify(simplify_geometry) for rp in self.bounds_poly]
        self._setup_boundaries()

    # def line_to_xy_boundary(self, line, buffer):
    #     return np.asarray(Polygon(LineString(line).buffer(buffer, join_style=2).exterior).exterior.coords)

    def get_xy_boundaries(self):
        polygons = []
        bounds = []
        for z in self.zones:
            if hasattr(z.dist2wt, '__code__'):
                buffer = z.dist2wt(**{k: 100 for k in z.dist2wt.__code__.co_varnames})
            else:
                buffer = 0
            if z.geometry_type == 'line':
                poly = Polygon(LineString(z.boundary).buffer(buffer, join_style=2).exterior)
            elif z.geometry_type == 'polygon':
                poly = Polygon(z.boundary).buffer(buffer, join_style=2)
            polygons.append(poly)
            bounds.append(np.asarray(poly.exterior.coords))
        return polygons, bounds

    def _setup_boundaries(self):
        self.res_poly = self._calc_resulting_polygons(self.bounds_poly)
        self.boundaries = self._poly_to_bound(self.res_poly)

        boundary_properties_list_all = list(zip(*[self.get_boundary_properties(bound, incl_excl)[1:]
                                                  for bound, incl_excl in self.boundaries]))

        self.boundary_properties_list_all = [np.concatenate(v, -1)
                                             for v in boundary_properties_list_all]

    def _poly_to_bound(self, polygons):
        boundaries = []
        for bound in polygons:
            x, y = bound.exterior.xy
            boundaries.append((np.asarray([x, y]).T[:-1, :], 1))
            for interior in bound.interiors:
                x, y = interior.xy
                boundaries.append((np.asarray([x, y]).T[:-1, :], 0))
        return boundaries

    def _calc_resulting_polygons(self, boundary_polygons):
        '''
        Parameters
        ----------
        boundary_polygons : list
            list of shapely polygons as specifed or inferred from user input
        Returns
        -------
        list of merged shapely polygons. Resolves issues arrising if any are overlapping, touching or contained in each other
        '''
        domain = []
        for i in range(len(boundary_polygons)):
            b = boundary_polygons[i]
            if len(domain) == 0:
                if self.incl_excls[i]:
                    domain.append(b)
                else:
                    warnings.warn("First boundary should be an inclusion zone or it will be ignored")
                    pass
            else:
                if self.incl_excls[i]:
                    temp = []
                    for j, d in enumerate(domain):
                        if d.intersects(b):
                            b = unary_union([d, b])
                        else:
                            if d.contains(b):
                                warnings.warn("Boundary is fully contained preceding polygon and will be ignored")
                                pass
                            elif b.contains(d):
                                b = d
                                warnings.warn("Boundary is fully containing preceding polygon and will override it")
                                pass
                            else:
                                if b.area > 1e-3:
                                    temp.append(d)
                        if j == len(domain) - 1:
                            if b.area > 1e-3:
                                temp.append(b)
                    domain = temp
                else:
                    temp = []
                    for j, d in enumerate(domain):
                        if d.intersects(b):
                            nonoverlap = (d.symmetric_difference(b)).difference(b)
                            if isinstance(nonoverlap, type(Polygon())):
                                temp.append(nonoverlap)
                            elif isinstance(nonoverlap, type(MultiPolygon())):
                                for x in nonoverlap.geoms:
                                    if x.area > 1e-3:
                                        temp.append(x)
                        else:
                            if b.contains(d):
                                warnings.warn("Exclusion boundary fully consumes preceding polygon")
                                pass
                            else:
                                if d.contains(b):
                                    d = Polygon(d.exterior.coords, [b.exterior.coords])
                                if d.area > 1e-3:
                                    temp.append(d)
                    domain = temp
        return domain

    def sign(self, Dist_ij):
        return np.sign(Dist_ij[np.arange(Dist_ij.shape[0]), np.argmin(abs(Dist_ij), axis=1)])

    def calc_distance_and_gradients(self, x, y):
        '''
        Parameters
        ----------
        x : 1d array
            Array of x-positions.
        y : 1d array
            Array of y-positions.

        Returns
        -------
        D_ij : 2d array
            Array of point-edge distances. index 'i' is points and index 'j' is total number of edges.
        sign_i : 1d array
            Array of signs of the governing distance.
        dDdk_jk : 2d array
            Jacobian of the distance matrix D_ij with respect to x and y.

        '''
        if np.all(np.array([x, y]) == self._cache_input) & (not self.relaxation):
            return self._cache_output

        Dist_ij, ddist_dX, ddist_dY = self._calc_distance_and_gradients(x, y, self.boundary_properties_list_all)

        dDdk_ijk = np.moveaxis([ddist_dX, ddist_dY], 0, -1)
        sign_i = self.sign(Dist_ij)
        self._cache_input = np.array([x, y])
        self._cache_output = [Dist_ij, dDdk_ijk, sign_i]
        return self._cache_output

    def calc_relaxation(self, iteration_no=None):
        '''
        The tupple relaxation contains a first term for the penalty constant
        and a second term for the n first iterations to apply relaxation.
        '''
        if iteration_no is None:
            iteration_no = self.problem.cost_comp.n_grad_eval + 1
        return max(0, self.relaxation[0] * (self.relaxation[1] - iteration_no))

    def distances(self, x, y):
        Dist_ij, _, sign_i = self.calc_distance_and_gradients(x, y)
        if self.method == 'smooth_min':
            Dist_i = smooth_max(np.abs(Dist_ij), -np.abs(Dist_ij).max(), axis=1) * sign_i
        elif self.method == 'nearest':
            Dist_i = Dist_ij[np.arange(x.size), np.argmin(np.abs(Dist_ij), axis=1)]
        else:
            warning = f'method: {self.method} is not implemented. Available options are smooth_min and nearest.'
            warnings.warn(warning)
        if self.relaxation:
            Dist_i += self.calc_relaxation()
        return Dist_i

    def gradients(self, x, y):
        '''
        The derivate of the smooth maximum with respect to x and y is calculated with the chain rule:
            dS/dk = dS/dD * dD/dk
            where S is smooth maximum, D is distance to edge and k is the spacial dimension
        '''
        Dist_ij, dDdk_ijk, _ = self.calc_distance_and_gradients(x, y)
        if self.relaxation:
            Dist_ij += self.calc_relaxation()
            # dDdt = -self.relaxation[1]
        if self.method == 'smooth_min':
            dSdDist_ij = smooth_max_gradient(np.abs(Dist_ij), -np.abs(Dist_ij).max(), axis=1)
            dSdkx_i, dSdky_i = (dSdDist_ij[:, :, na] * dDdk_ijk).sum(axis=1).T
        elif self.method == 'nearest':
            dSdkx_i, dSdky_i = dDdk_ijk[np.arange(x.size), np.argmin(np.abs(Dist_ij), axis=1), :].T

        if self.relaxation:
            # as relaxed distance is relaxation + distance, the gradient with respect to x and y is unchanged
            gradients = np.diagflat(dSdkx_i), np.diagflat(dSdky_i), np.ones(self.n_wt) * self.relaxation[1]
        else:
            gradients = np.diagflat(dSdkx_i), np.diagflat(dSdky_i)
        return gradients

    def relaxed_polygons(self, iteration_no=None):
        poly = [Polygon(x.boundary) for x in self.zones]
        booleans = [x.incl for x in self.zones]
        relaxed_poly = []
        for i, p in enumerate(poly):
            if booleans[i] == 0:
                pb = p.buffer(-self.calc_relaxation(iteration_no), join_style=2)
                relaxed_poly.append(pb)
            else:
                pb = p.buffer(self.calc_relaxation(iteration_no), join_style=2)
                relaxed_poly.append(pb)
        merged_poly = self._calc_resulting_polygons(relaxed_poly)
        return self._poly_to_bound(merged_poly)


class TurbineSpecificBoundaryComp(MultiPolygonBoundaryComp):
    def __init__(self, n_wt, wind_turbines, zones, const_id=None,
                 units=None, relaxation=False, method='nearest', simplify_geometry=False):
        # self.dependencies = [d or {'type': None, 'multiplier': None, 'ref': None} for d in dependencies]
        # self.multi_boundary = xy_boundaries = self.get_xy_boundaries(boundaries, geometry_types, incl_excls)
        self.wind_turbines = wind_turbines
        self.types = wind_turbines.types()
        # self.incl_excls = incl_excls
        self.n_wt = n_wt
        self.zones = zones
        self.ts_polygon_boundaries, ts_xy_boundaries = self.get_ts_boundaries()
        MultiPolygonBoundaryComp.__init__(self, n_wt=n_wt, zones=zones, const_id=const_id, units=units,
                                          relaxation=relaxation, method=method, simplify_geometry=simplify_geometry)
        # self.polygon_boundaries = [Polygon(x) for x, _ in xy_boundaries]
        # self.ts_polygon_boundaries = self.get_ts_polygon_boundaries(self.types)
        self.ts_merged_polygon_boundaries = self.merge_boundaries()
        self.ts_merged_xy_boundaries = self.get_ts_xy_boundaries()
        self.ts_boundary_properties = self.get_ts_boundary_properties()
        self.ts_item_indices = self.get_ts_item_indices()

    def get_ts_boundaries(self):
        polygons = []
        bounds = []
        for t in set(self.types):
            temp1 = []
            temp2 = []
            dist2wt_input = dict(D=self.wind_turbines.diameter(t),
                                 H=self.wind_turbines.hub_height(t))
            for z in self.zones:
                if hasattr(z.dist2wt, '__code__'):
                    buffer = z.dist2wt(**{k: dist2wt_input[k] for k in z.dist2wt.__code__.co_varnames})
                else:
                    buffer = 0
                if z.geometry_type == 'line':
                    poly = Polygon(LineString(z.boundary).buffer(buffer, join_style=2).exterior)
                elif z.geometry_type == 'polygon':
                    poly = Polygon(z.boundary).buffer(buffer, join_style=2)
                bound = np.asarray(poly)
                temp1.append(poly)
                temp2.append(bound)
            polygons.append(temp1)
            bounds.append(temp2)
        return polygons, bounds
        # for wt
        # temp = []
        # for n, (b, t, ie) in enumerate(zip(boundaries, geometry_types, incl_excls)):
        #     if t == 'line':
        #         bound = np.asarray(Polygon(LineString(b).buffer(default_ref, join_style=2).exterior).exterior.coords)
        #         self.dependencies[n]['ref'] = default_ref
        #     elif t == 'polygon':
        #         bound = b
        #     else:
        #         raise NotImplementedError("Geometry type '%s' is not implemented" % b)
        #     temp.append((bound, ie))

    # def get_ts_polygon_boundaries(self, types):
    #     temp = []
    #     for t in set(types):
    #         d = self.wind_turbines.diameter(t)
    #         h = self.wind_turbines.hub_height(t)
    #         temp.append(self.get_ts_polygon_boundary(d, h))
    #     return temp

    def get_ts_xy_boundaries(self):
        return [self._poly_to_bound(b) for b in self.ts_merged_polygon_boundaries]

    # def get_ts_polygon_boundary(self, d=None, h=None):
    #     temp = []
    #     for bound, dep in zip(self.polygon_boundaries, self.dependencies):
    #         ref = dep['ref'] or 0
    #         if dep['type'] == 'D':
    #             ts_polygon_boundary = bound.buffer(dep['multiplier']*d-ref, join_style=2)
    #         elif dep['type'] == 'H':
    #             ts_polygon_boundary = bound.buffer(dep['multiplier']*h-ref, join_style=2)
    #         else:
    #             ts_polygon_boundary = bound
    #         temp.append(ts_polygon_boundary)
    #     return temp

    def merge_boundaries(self):
        return [self._calc_resulting_polygons(bounds) for bounds in self.ts_polygon_boundaries]

    def get_ts_boundary_properties(self,):
        return [[self.get_boundary_properties(bound) for bound, _ in bounds] for bounds in self.ts_merged_xy_boundaries]

    def get_ts_item_indices(self):
        temp = []
        for bounds in self.ts_merged_xy_boundaries:
            n_edges = np.asarray([len(bound) for bound, _ in bounds])
            n_edges_tot = np.sum(n_edges)
            start_at = np.cumsum(n_edges) - n_edges
            end_at = start_at + n_edges
            item_indices = [n_edges_tot, start_at, end_at]
            temp.append(item_indices)
        return temp

    def calc_distance_and_gradients(self, x, y, types=None):
        if not isinstance(self._cache_input, type(None)):
            if np.all(np.array([x, y]) == self._cache_input[0]) & (not self.relaxation) & np.all(np.asarray([types]) == self._cache_input[1]):
                return self._cache_output
        if isinstance(types, type(None)):
            types = np.zeros(self.n_wt)
        Dist_i = np.zeros(self.n_wt)
        sign_i = np.zeros(self.n_wt)
        dDdx_i = np.zeros(self.n_wt)
        dDdy_i = np.zeros(self.n_wt)
        for t in set(types):
            t = int(t)
            idx = (types == t)
            n_edges_tot, start_at, end_at = self.ts_item_indices[t]
            Dist_ij = np.zeros((sum(idx), n_edges_tot))
            dDdk_ijk = np.zeros((sum(idx), n_edges_tot, 2))
            for n, (bound, bound_type) in enumerate(self.ts_merged_xy_boundaries[t]):
                sa = start_at[n]
                ea = end_at[n]
                distance, ddist_dX, ddist_dY = self._calc_distance_and_gradients(x[idx], y[idx], self.ts_boundary_properties[t][n][1:])
                if bound_type == 0:
                    distance *= -1
                    ddist_dX *= -1
                    ddist_dY *= -1
                Dist_ij[:, sa:ea] = distance
                dDdk_ijk[:, sa:ea, 0] = ddist_dX
                dDdk_ijk[:, sa:ea, 1] = ddist_dY

            sign_i[idx] = self.sign(Dist_ij)
            Dist_i[idx] = Dist_ij[np.arange(sum(idx)), np.argmin(np.abs(Dist_ij), axis=1)]
            dDdx_i[idx], dDdy_i[idx] = dDdk_ijk[np.arange(sum(idx)), np.argmin(np.abs(Dist_ij), axis=1), :].T
        self._cache_input = (np.array([x, y]), np.asarray(types))
        self._cache_output = [Dist_i, dDdx_i, dDdy_i, sign_i]
        return self._cache_output

    def distances(self, x, y, type=None):
        Dist_i, _, _, _ = self.calc_distance_and_gradients(x, y, types=type)
        if self.relaxation:
            Dist_i += self.calc_relaxation()
        return Dist_i

    def gradients(self, x, y, type=None):
        Dist_i, dDdx_i, dDdy_i, _ = self.calc_distance_and_gradients(x, y, types=type)
        if self.relaxation:
            Dist_i += self.calc_relaxation()
            dDdt = -self.relaxation[0]
        if self.relaxation:
            gradients = np.diagflat(dDdx_i), np.diagflat(dDdy_i), np.ones(self.n_wt) * dDdt
        else:
            gradients = np.diagflat(dDdx_i), np.diagflat(dDdy_i)
        return gradients


def main():
    if __name__ == '__main__':
        import matplotlib.pyplot as plt
        from py_wake.wind_turbines import WindTurbines
        from py_wake.wind_turbines.power_ct_functions import CubePowerSimpleCt

        plt.close('all')
        i1 = np.array([[2, 17], [6, 23], [16, 23], [26, 15], [19, 0], [14, 4], [4, 4]])
        e1 = np.array([[0, 10], [20, 21], [22, 12], [10, 12], [9, 6], [2, 7]])
        i2 = np.array([[12, 13], [14, 17], [18, 15], [17, 10], [15, 11]])
        e2 = np.array([[5, 17], [5, 18], [8, 19], [8, 18]])
        i3 = np.array([[5, 0], [5, 1], [10, 3], [10, 0]])
        e3 = np.array([[6, -1], [6, 18], [7, 18], [7, -1]])
        e4 = np.array([[15, 9], [15, 11], [20, 11], [20, 9]])
        e5 = np.array([[10, 25], [20, 0]])
        zones = [
            InclusionZone(i1, name='i1'),
            InclusionZone(i2, name='i2'),
            InclusionZone(i3, name='i3'),
            ExclusionZone(e1, name='e1'),
            ExclusionZone(e2, name='e2'),
            ExclusionZone(e3, name='e3'),
            ExclusionZone(e4, name='e4'),
            ExclusionZone(e5, name='e5', dist2wt=lambda: 1, geometry_type='line'),
        ]

        N_points = 50
        xs = np.linspace(-1, 30, N_points)
        ys = np.linspace(-1, 30, N_points)
        y_grid, x_grid = np.meshgrid(xs, ys)
        x = x_grid.ravel()
        y = y_grid.ravel()
        n_wt = len(x)
        MPBC = MultiPolygonBoundaryComp(n_wt, zones, method='nearest')
        distances = MPBC.distances(x, y)
        delta = 1e-9
        distances2 = MPBC.distances(x + delta, y)
        dx_fd = (distances2 - distances) / delta
        dx = np.diag(MPBC.gradients(x + delta / 2, y)[0])

        plt.figure()
        plt.plot(dx_fd, dx, '.')

        plt.figure()
        for n, bound in enumerate(MPBC.boundaries):
            x_bound, y_bound = bound[0].T
            x_bound = np.append(x_bound, x_bound[0])
            y_bound = np.append(y_bound, y_bound[0])
            line, = plt.plot(x_bound, y_bound, label=f'{n}')
            plt.plot(x_bound[0], y_bound[0], color=line.get_color(), marker='o')

        plt.legend()
        plt.grid()
        plt.axis('square')
        plt.contourf(x_grid, y_grid, distances.reshape(N_points, N_points), np.linspace(-10, 10, 100), cmap='seismic')
        plt.colorbar()

        plt.figure()
        ax = plt.axes(projection='3d')
        ax.contour3D(
            x.reshape(
                N_points, N_points), y.reshape(
                N_points, N_points), distances.reshape(
                N_points, N_points), np.linspace(-10, 10, 100), cmap='seismic')
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')

        if 0:
            for smpl in [0, 1, 2, 3, 4, 5, 6, 7, 8]:
                MPBC = MultiPolygonBoundaryComp(n_wt, zones, simplify_geometry=smpl)
                plt.figure()
                ax = plt.gca()
                MPBC.plot(ax)

        wind_turbines = WindTurbines(names=['tb1', 'tb2'],
                                     diameters=[80, 120],
                                     hub_heights=[70, 110],
                                     powerCtFunctions=[
            CubePowerSimpleCt(ws_cutin=3, ws_cutout=25, ws_rated=12,
                              power_rated=2000, power_unit='kW',
                              ct=8 / 9, additional_models=[]),
            CubePowerSimpleCt(ws_cutin=3, ws_cutout=25, ws_rated=12,
                              power_rated=3000, power_unit='kW',
                              ct=8 / 9, additional_models=[])])

        x1 = [0, 3000, 3000, 0]
        y1 = [0, 0, 3000, 3000]
        b1 = np.transpose((x1, y1))

        # Buildings
        x2 = [600, 1400, 1400, 600]
        y2 = [1700, 1700, 2500, 2500]
        b2 = np.transpose((x2, y2))

        # River
        x3 = np.linspace(520, 2420, 16)
        y3 = [0, 133, 266, 400, 500, 600, 700, 733, 866, 1300, 1633,
              2100, 2400, 2533, 2700, 3000]
        b3 = np.transpose((x3, y3))

        # Roads
        x4 = np.linspace(0, 3000, 16)
        y4 = [1095, 1038, 1110, 1006, 1028, 992, 977, 1052, 1076, 1064, 1073,
              1027, 964, 981, 1015, 1058]
        b4 = np.transpose((x4, y4))

        zones = [
            InclusionZone(b1, name='i1'),
            ExclusionZone(b2, dist2wt=lambda H: 4 * H - 360, name='building'),
            ExclusionZone(b3, geometry_type='line', dist2wt=lambda D: 3 * D, name='river'),
            ExclusionZone(b4, geometry_type='line', dist2wt=lambda D, H: max(D * 2, H * 3), name='road'),
        ]
        N_points = 50
        xs = np.linspace(0, 3000, N_points)
        ys = np.linspace(0, 3000, N_points)
        y_grid, x_grid = np.meshgrid(xs, ys)
        x = x_grid.ravel()
        y = y_grid.ravel()
        n_wt = len(x)
        types = np.zeros(n_wt)
        TSBC = TurbineSpecificBoundaryComp(n_wt, wind_turbines, zones)
        distances = TSBC.distances(x, y, type=types)
        delta = 1e-9
        distances2 = TSBC.distances(x + delta, y, type=types)
        dx_fd = (distances2 - distances) / delta
        dx = np.diag(TSBC.gradients(x + delta / 2, y, type=types)[0])

        plt.figure()
        plt.plot(dx_fd, dx, '.')

        plt.figure()
        for ll, t in enumerate(TSBC.types):
            line, = plt.plot(*TSBC.ts_merged_xy_boundaries[ll][0][0][0, :], label=f'type {ll}')
            for n, bound in enumerate(TSBC.ts_merged_xy_boundaries[ll]):
                x_bound, y_bound = bound[0].T
                x_bound = np.append(x_bound, x_bound[0])
                y_bound = np.append(y_bound, y_bound[0])
                plt.plot(x_bound, y_bound, color=line.get_color())

        plt.legend()
        plt.grid()
        plt.axis('square')

        for ll, t in enumerate(TSBC.types):
            plt.figure()
            for n, bound in enumerate(TSBC.ts_merged_xy_boundaries[ll]):
                x_bound, y_bound = bound[0].T
                x_bound = np.append(x_bound, x_bound[0])
                y_bound = np.append(y_bound, y_bound[0])
                plt.plot(x_bound, y_bound, 'b')
            plt.grid()
            plt.title(f'type {ll}')
            plt.axis('square')
            plt.contourf(x_grid, y_grid, TSBC.distances(x, y, type=t * np.ones(n_wt)).reshape(N_points, N_points), 50)
            plt.colorbar()


main()



# -*- coding: utf-8 -*-
"""
Setup file for Topfarm2
"""


import os
from git_utils import write_vers
from setuptools import setup, find_packages

repo = os.path.dirname(__file__)
version = write_vers(vers_file='topfarm/__init__.py', repo=repo, skip_chars=1)


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(name='topfarm',
      version=version,
      description='Topfarm - Wind farm optimization using OpenMDAO',
      long_description=read('README'),
      url='https://gitlab.windenergy.dtu.dk/TOPFARM/topfarm2',
      author='DTU Wind Energy',
      author_email='mikf@dtu.dk',
      license='MIT',
      packages=find_packages(),
      install_requires=[
              'matplotlib',  # for plotting
              'numpy',  # for numerical calculations
              'numpy-financial',  # for irr calculations
              'openmdao>=3.16',  # for optimization
              'pytest',  # for testing
              'pycodestyle',  # for testing
              'pytest-cov',  # for calculating coverage
              'py_wake>=2',  # for calculating AEP
              'pyDOE2',  # for GA driver
              'pymongo[srv]',  # for mongo_recorder
              'scipy',  # constraints
              'sphinx',  # generating documentation
              'sphinx_rtd_theme',  # docs theme
              'scikit-learn',  # load surrogate
              'mock',  # replace variables during tests
              'tensorflow',  # loads examples with surrogates
              'shapely>=1.8.2',  # for multiple polygon boundaries and exclusion zones
              'memory_profiler',  # memory profiling
      ],
      zip_safe=True)



import numpy as np
import matplotlib.pyplot as plt
from numpy import newaxis as na
import multiprocessing
import threading
import time
from tqdm import tqdm
from openmdao.core.explicitcomponent import ExplicitComponent
import topfarm
from abc import abstractmethod, ABC
import gc


def smart_start(XX, YY, ZZ, N_WT, min_space, radius=None, random_pct=0, plot=False, seed=None, types=None):
    """Selects the a number of gridpoints (N_WT) in the grid defined by x and y,
    where ZZ has the maximum value, while chosen points spacing (min_space)
    is respected.

    Parameters
    ----------
    XX : array_like
        x coordinates
    YY : array_like
        y coordinates
    ZZ : array_like
        Values at (XX, YY) of the desired variable in the
        grid points. This could be e.g. the AEP or wind speed.
    N_WT : integer
        number of wind turbines
    min_space: float
        minimum space between turbines
    random_pct : float
        select by random position of the <random_pct> best points
    plot : boolean
        if True, each step is plotted in new figure
    types : array_like of integers or None
        list of turbine type numbers e.g. types = [0, 1, 2, 3] for 4 different types. ZZ, min_space and optionally
        also radius should also have type dimension. if ZZ is callable it should be callable with type argument.

    Returns
    -------
    Positions where the aep or wsp is highest, while
    respecting the minimum spacing requirement.

    Notes
    -----
    XX, YY and ZZ can be 1D or 2D, but must have same size. If multiple turbine types
    ZZ must have the shape (n_types, shape_like_XX)
    """
    assert 0 <= random_pct <= 100
    ZZ_is_func = hasattr(ZZ, '__call__')
    if types:
        n_types = len(types)
        n_points = XX.size
        arr = np.zeros((3, n_points, n_types))
        arr[0, :, :] = XX.ravel()[:, np.newaxis]
        arr[1, :, :] = YY.ravel()[:, np.newaxis]
        if ZZ_is_func:
            arr[2, :, :] = np.zeros_like(arr[0])
        else:
            arr[2, :, :] = ZZ.reshape(n_types, -1).T
    else:
        if ZZ_is_func:
            arr = np.array([XX.flatten(), YY.flatten()])
        else:
            arr = np.array([XX.flatten(), YY.flatten(), ZZ.flatten()])

    # set radius to None(faster) if if grid resolution > radius
    if radius is not None and (np.diff(np.sort(np.unique(arr[0]))).min() > radius and
                               (np.diff(np.sort(np.unique(arr[1]))).min() > radius)):
        radius = None
    xs, ys = [], []
    if types:
        type_i = []
    if seed is None:
        seed = np.uint32(int((time.time() - int(time.time())) * 1e8 +
                             multiprocessing.current_process().ident + threading.get_ident()) % (2**31))
        np.random.seed(seed)
        seed = np.random.randint(0, 2**31)
    np.random.seed(seed)
    mask = np.full(arr.shape[1:], True)
    for i in tqdm(range(N_WT), desc='Smartstart'):
        idx = np.where(mask < 1)
        if arr.shape[1] == 0:
            raise Exception('No feasible positions for wt %d' % i)

        if ZZ_is_func:
            if random_pct < 100:
                if types:
                    z = np.asarray([ZZ(arr[0, :, tt], arr[1, :, tt], xs, ys, tt, type_i) for tt in types]).T
                    z[idx] = -np.inf
                    arr[2] = z
                else:
                    z = ZZ(arr[0], arr[1], xs, ys)
            else:
                z = np.zeros_like(arr[0])
                if types:
                    arr[2] = z
        else:
            z = arr[2]
            if types:
                z[idx] = -np.inf

        if radius is not None:
            # average over the rotor area, i.e. all points within one radius from the point
            x, y = arr
            z = np.array([np.mean(z[ind]) for ind in np.hypot((x - x[:, na]), (y - y[:, na])) < radius])

        # pick one of the <random_pct> best points
        n_random = np.maximum(1, int(np.round(random_pct / 100 * z.size)))
        min_z = np.sort(z.ravel())[-(n_random)]
        if types:
            choises = np.argwhere(z >= min_z)
            arg_choise = np.random.choice(np.arange(choises.shape[0]))
            next_ind, t = choises[arg_choise]
            next_ind = (next_ind, 0)
            type_i.append(t)
        else:
            next_ind = np.random.choice(np.argwhere(z >= min_z)[:, 0])

        x0 = arr[0][next_ind]
        y0 = arr[1][next_ind]
        xs.append(x0)
        ys.append(y0)

        if plot:
            if types:
                for typ in types:
                    plt.figure()
                    c = plt.scatter(arr[0, :, typ], arr[1, :, typ], c=arr[2, :, typ])
                    plt.colorbar(c)
                    plt.plot(xs, ys, '2k', ms=10)
                    plt.plot(xs[-1], ys[-1], '2r', ms=10)
                    plt.axis('equal')
                    plt.show()
            else:
                plt.figure()
                c = plt.scatter(arr[0], arr[1], c=z)
                plt.colorbar(c)
                plt.plot(xs, ys, '2k', ms=10)
                plt.plot(xs[-1], ys[-1], '2r', ms=10)
                plt.axis('equal')
                plt.show()

        # Remove all point within min_space from the newly added wt
        if types:
            eff_min_space = (min_space + min_space[t]) / 2
            mask = np.logical_and(mask, ((arr[0, :, 0] - x0) ** 2 + (arr[1, :, 0] - y0) ** 2)[:, na] >= eff_min_space[na, :] ** 2)
            idx = mask.sum((1)) > 0
            arr = arr[:, idx, :]
            mask = mask[idx, :]
        else:
            index = np.where((arr[0] - x0)**2 + (arr[1] - y0)**2 >= min_space**2)[0]
            arr = arr[:, index]
        gc.collect()

    print(
        f"{len(XX.flatten())} possible points, {N_WT} wt, {len(XX)/N_WT:.1f} points pr wt, {arr.shape[1]}({arr.shape[1]/len(XX.flatten())*100:.0f}%) unused points")
    if types:
        return xs, ys, type_i
    else:
        return xs, ys


def smooth_max(X, alpha, axis=0):
    '''
    Returns the smooth maximum of a matrix for positive values of alpha and smoth minimum for negative values of alpha
    Parameters
    ----------
    X : ndarray
        Matrix of which the smooth maximum is calculated.
    alpha : float
        smoothness parameter.
    axis : int, optional
        Axis along which the smooth maximum is calculated. The default is 0.

    Returns
    -------
    ndarray
        Matrix of smooth maximum values.

    '''
    return (X * np.exp(alpha * X)).sum(axis=axis) / np.exp(alpha * X).sum(axis=axis)


def smooth_max_gradient(X, alpha, axis=0):
    '''
    Parameters
    ----------
    X : ndarray
        Matrix of which the smooth maximum derivative is calculated.
    alpha : float
        smoothness parameter.
    axis : int, optional
        Axis along which the smooth maximum is calculated. The default is 0. The default is 0.

    Returns
    -------
    ndarray
        Matrix of smooth maximum derivatives.

    '''
    return np.exp(alpha * X) / np.expand_dims(np.exp(alpha * X).sum(axis=axis), axis) * \
        (1 + alpha * (X - np.expand_dims(smooth_max(X, alpha, axis=axis), axis)))


def gauss(X):
    return np.exp(-X ** 2)


def gauss_inv(X):
    return np.sqrt(-np.log(X))


def gauss_prime(X):
    return -2 * X * gauss(X)


def gauss_inv_prime(X):
    return -1 / (2 * X * gauss_inv(X))


def smooth_zero(X, alpha=1, axis=0):
    '''
    Function that from an array approximates the value cloest to zero. A differentiable alternative
    to min(abs(X)).
    Parameters
    ----------
    X : ndarray
        Matrix of which the smooth zero is calculated.
    alpha : float
        smoothness parameter.
    axis : int, optional
        Axis along which the smooth zero is calculated. The default is 0.

    Returns
    -------
    ndarray
        Matrix of smooth zero.

    '''
    sign = np.sign(X[np.expand_dims(np.argmin(np.abs(X)), axis)])
    return sign * gauss_inv(smooth_max(gauss(X), alpha, axis))


def smooth_zero_gradient(X, alpha=1, axis=0):
    '''
    Gradients of the smooth_zero function.
    Parameters
    ----------
    X : ndarray
        Matrix of which the smooth zero derivative is calculated.
    alpha : float
        smoothness parameter.
    axis : int, optional
        Axis along which the smooth zero is calculated. The default is 0.

    Returns
    -------
    ndarray
        Matrix of smooth zero derivatives.

    '''
    sign = np.sign(X[np.expand_dims(np.argmin(np.abs(X)), axis)])
    return sign * gauss_prime(X) * smooth_max_gradient(gauss(X), alpha, axis) * gauss_inv_prime(smooth_max(gauss(X), alpha, axis))


class AggregationFunction(ABC):
    def __str__(self):
        return self.__class__.__name__

    @abstractmethod
    def __call__(self, x, axis=-1):
        """compute function value over axis"""

    @abstractmethod
    def gradient(self, x, axis=-1):
        """compute gradients of aggregated value wrt x over axis"""


class StrictMax(AggregationFunction):
    """Normal max with discontinous gradient"""

    def __call__(self, x, axis=-1):
        return np.max(x, axis)

    def gradient(self, x, axis=-1):
        return x == self(np.atleast_2d(x), axis)


class StrictMin(StrictMax):
    """Normal min with discontinous gradient"""

    def __call__(self, x, axis=-1):
        return np.min(x, axis)


def SoftMax(x, alpha, axis=-1):
    """gradient of LogSumExpMax

    https://en.wikipedia.org/wiki/Softmax_function"""
    if alpha > 0:
        x_max = np.max(np.atleast_2d(x), axis)
    else:
        x_max = np.min(np.atleast_2d(x), axis)
    xn = x - np.expand_dims(x_max, axis)
    return np.exp(alpha * xn) / np.sum(np.exp(alpha * xn), axis)


class SmoothMax(AggregationFunction):
    def __init__(self, base=1):
        """
        https://en.wikipedia.org/wiki/Smooth_maximum

        Underpredict the maximum of similar values

        Parameters
        ----------
        base : float
            smoothing factor ]0;inf[. higher number gives more smooth transition
            For two numbers, (a,b), the smoothing approximately starts when |a-b| < 4*base
        """
        self.alpha = 1 / base

    def __str__(self):
        return f"{self.__class__.__name__}({1/self.alpha})"

    def __call__(self, x, axis=-1):
        if self.alpha > 0:
            x_max = np.max(np.atleast_2d(x), axis)
        else:
            x_max = np.min(np.atleast_2d(x), axis)
        xn = x - np.expand_dims(x_max, axis)
        return np.sum(x * np.exp(self.alpha * xn), axis) / np.sum(np.exp(self.alpha * xn), axis)

    def gradient(self, x, axis=-1):
        x = np.asarray(x)
        a = self.alpha
        if a > 0:
            x_max = np.max(np.atleast_2d(x), axis)
        else:
            x_max = np.min(np.atleast_2d(x), axis)
        xn = x - np.expand_dims(x_max, axis)
        return (np.exp(a * xn) /
                np.expand_dims(np.sum(np.exp(a * xn), axis), axis) * (1 + a * (x - np.expand_dims(self(x, axis), axis))))


class SmoothMin(SmoothMax):
    def __init__(self, base=1):
        SmoothMax.__init__(self, base=-base)

    def __str__(self):
        return f"{self.__class__.__name__}({-1/self.alpha})"


class LogSumExpMax(SmoothMax):
    """LogSumExp

    Overpredict the maximum of similar values

    Parameters
    ----------
    base : float
        smoothing factor ]0;inf[. higher number gives more smooth transition
        For two numbers, (a,b), the smoothing approximately starts when |a-b| < 4*base
    """

    def __str__(self):
        return f"{self.__class__.__name__}({1/self.alpha})"

    def __call__(self, x, axis=-1):
        # factor used to reduce numerical errors in power
        if self.alpha > 0:
            x_max = np.max(np.atleast_2d(x), axis)
        else:
            x_max = np.min(np.atleast_2d(x), axis)
        xn = x - np.expand_dims(x_max, axis)
        alpha = self.alpha
        return x_max + 1 / alpha * np.log(np.sum(np.exp(alpha * (xn)), axis))

    def gradient(self, x, axis=-1):
        return SoftMax(x, self.alpha, axis)


class LogSumExpMin(LogSumExpMax):
    def __init__(self, base=1):
        LogSumExpMax.__init__(self, base=-base)

    def __str__(self):
        return f"{self.__class__.__name__}({-1/self.alpha})"


def regular_generic_layout(n_wt, sx, sy, stagger, rotation, x0=0, y0=0, ratio=1.0):
    '''
    Parameters
    ----------
    n_wt : int
        number of wind turbines
    sx : float
        spacing (in turbine diameters or meters) between turbines in x direction
    sy : float
        spacing (in turbine diameters or meters) between turbines in y direction
    stagger : float
        stagger (in turbine diameters or meters) distance every other turbine column
    rotation : float
        rotational angle of the grid in degrees
    ratio : float
        ratio between number of columns and number of rows (1.0)

    Returns
    -------
    xy : array
        2D array of x- and y-coordinates (in turbine diameters or meters)

     '''

    if ratio > 1:
        n_row = int(np.round((n_wt * ratio) ** 0.5))
        n_col = int(n_wt / n_row)
    else:
        n_col = int(np.round((n_wt * ratio) ** 0.5))
        n_row = int(n_wt / n_col)
    rest = n_wt - n_col * n_row
    theta = np.radians(float(rotation))
    R = np.array([[np.cos(theta), -np.sin(theta)],
                  [np.sin(theta), np.cos(theta)]])
    x_grid = np.linspace(0, n_col * float(sx), n_col, endpoint=False)
    y_grid = np.linspace(0, n_row * float(sy), n_row, endpoint=False)
    xm, ym = np.meshgrid(x_grid, y_grid)
    ym[:, 1::2] += stagger
    x, y = xm.ravel(), ym.ravel()
    x = np.hstack((x, x_grid[:rest]))
    y = np.hstack((y, (y[-n_col:-(n_col - rest)] + float(sy))))
    xy_grid = np.matmul(R, np.array([x, y]))
    return xy_grid + np.asarray([x0, y0])[:, na]


def regular_generic_layout_gradients(n_wt, sx, sy, stagger, rotation, x0=0, y0=0, ratio=1.0):
    '''
    Parameters
    ----------
    n_wt : int
        number of wind turbines
    sx : float
        spacing (in turbine diameters or meters) between turbines in x direction
    sy : float
        spacing (in turbine diameters or meters) between turbines in y direction
    stagger : float
        stagger (in turbine diameters or meters) distance every other turbine column
    rotation : float
        rotational angle of the grid in degrees
    ratio : float
        ratio between number of columns and number of rows (1.0)

    Returns
    -------
    dx_dsx, dy_dsx, dx_dsy, dy_dsy, dx_dr, dy_dr : tuple
        tuple of gradients of x and y with respect to x-spacing, y-spacing and grid rotation

     '''

    if ratio > 1:
        n_row = int(np.round((n_wt * ratio) ** 0.5))
        n_col = int(n_wt / n_row)
    else:
        n_col = int(np.round((n_wt * ratio) ** 0.5))
        n_row = int(n_wt / n_col)
    rest = n_wt - n_col * n_row
    theta = np.radians(float(rotation))
    R = np.array([[np.cos(theta), -np.sin(theta)],
                  [np.sin(theta), np.cos(theta)]])

    x_grid = np.linspace(0, n_col, n_col, endpoint=False)
    y_grid = np.zeros(n_row)
    xm, ym = np.meshgrid(x_grid, y_grid)
    x, y = xm.ravel(), ym.ravel()
    x = np.hstack((x, x_grid[:rest]))
    y = np.hstack((y, (y[-n_col:-(n_col - rest)])))
    dxy_dsx = np.matmul(R, np.array([x, y]))
    dx_dsx = dxy_dsx[0, :]
    dy_dsx = dxy_dsx[1, :]

    x_grid = np.zeros(n_col)
    y_grid = np.linspace(0, n_row, n_row, endpoint=False)
    xm, ym = np.meshgrid(x_grid, y_grid)
    x, y = xm.ravel(), ym.ravel()
    x = np.hstack((x, x_grid[:rest]))
    y = np.hstack((y, (y[-n_col:-(n_col - rest)] + 1)))
    dxy_dsy = np.matmul(R, np.array([x, y]))
    dx_dsy = dxy_dsy[0, :]
    dy_dsy = dxy_dsy[1, :]

    x_grid = np.linspace(0, n_col * float(sx), n_col, endpoint=False)
    y_grid = np.linspace(0, n_row * float(sy), n_row, endpoint=False)
    xm, ym = np.meshgrid(x_grid, y_grid)
    ym[:, 1::2] += stagger
    x, y = xm.ravel(), ym.ravel()
    x = np.hstack((x, x_grid[:rest]))
    y = np.hstack((y, (y[-n_col:-(n_col - rest)] + float(sy))))
    dRdr = np.array([[-np.sin(theta), -np.cos(theta)],
                     [np.cos(theta), -np.sin(theta)]])
    dx_dr, dy_dr = np.matmul(dRdr, np.array([x, y])) * np.pi / 180

    return [dx_dsx, dy_dsx, dx_dsy, dy_dsy, dx_dr, dy_dr]


def main():
    if __name__ == '__main__':
        plt.close('all')
        N_WT = 30
        min_space = 2.1

        x = np.arange(0, 20, 0.1)
        y = np.arange(0, 10, 0.1)
        YY, XX = np.meshgrid(y, x)
        val = np.sin(XX) + np.sin(YY)
        min_space = 2.1
        xs, ys = smart_start(XX, YY, val, N_WT, min_space)
        plt.figure()
        c = plt.contourf(XX, YY, val, 100)
        plt.colorbar(c)
        for i in range(N_WT):
            circle = plt.Circle((xs[i], ys[i]), min_space / 2, color='b', fill=False)
            plt.gcf().gca().add_artist(circle)
            plt.plot(xs[i], ys[i], 'rx')
        plt.axis('equal')
        plt.show()

        plt.figure()
        xy = regular_generic_layout(n_wt=100, sx=5, sy=4, stagger=2, rotation=35, ratio=1.25, x0=20, y0=40)

        dsx = 0.1
        dxyx = regular_generic_layout(n_wt=100, sx=5 + dsx, sy=4, stagger=2, rotation=35, ratio=1.25, x0=20, y0=40)
        dxy_dsx = (dxyx - xy) / dsx
        dx_dsx, dy_dsx = dxy_dsx

        dsy = 0.1
        dxyy = regular_generic_layout(n_wt=100, sx=5, sy=4 + dsy, stagger=2, rotation=35, ratio=1.25, x0=20, y0=40)
        dxy_dsy = (dxyy - xy) / dsy
        dx_dsy, dy_dsy = dxy_dsy

        dR = 0.1
        dxyR = regular_generic_layout(n_wt=100, sx=5, sy=4, stagger=2, rotation=35 + dR, ratio=1.25, x0=20, y0=40)
        dxy_dR = (dxyR - xy) / dR
        dx_dR, dy_dR = dxy_dR

        x, y = xy
        plt.plot(x, y, '.')
        plt.axis('equal')

        dx_dsxg, dy_dsxg, dx_dsyg, dy_dsyg, dx_dRg, dy_dRg = regular_generic_layout_gradients(
            n_wt=100, sx=5, sy=4, stagger=2, rotation=35, ratio=1.25, x0=20, y0=40)
        plt.figure()
        plt.plot(dx_dsx.ravel(), dx_dsxg.ravel(), '.')
        plt.figure()
        plt.plot(dy_dsx.ravel(), dy_dsxg.ravel(), '.')
        plt.figure()
        plt.plot(dx_dsy.ravel(), dx_dsyg.ravel(), '.')
        plt.figure()
        plt.plot(dy_dsy.ravel(), dy_dsyg.ravel(), '.')

        plt.figure()
        plt.plot(dx_dR.ravel(), dx_dRg.ravel(), '.')
        plt.figure()
        plt.plot(dy_dR.ravel(), dy_dRg.ravel(), '.')


main()



import os

import pytest

from topfarm.tests.notebook import Notebook
import topfarm


def get_notebooks():
    def get(path):
        return [Notebook(path + f) for f in [f for f in os.listdir(path) if f.endswith('.ipynb')]]
    path = os.path.dirname(topfarm.__file__) + "/../docs/notebooks/"
    return get(path)


notebooks = get_notebooks()


@pytest.mark.parametrize("notebook", notebooks, ids=[os.path.basename(nb.filename) for nb in notebooks])
def test_notebooks(notebook):
    if os.path.basename(notebook.filename) in [
            # 'bathymetry.ipynb',  # ok
            # 'constraints.ipynb',  # ok
            # 'cost_models.ipynb',  # ok
            # 'drivers.ipynb',  #  ok
            # 'exclusion_zones.ipynb',  # ok
            'layout_and_loads.ipynb',  # gives error from tensorflow on synnefo machine
            # 'problems.ipynb',  # ok
            'roads_and_cables.ipynb',  # fails
            'wake_steering_and_loads.ipynb']:  # ok but many warnings from tensorflow
        pytest.xfail("Notebook, %s, has known issues" % notebook)
    import matplotlib.pyplot as plt

    def no_show(*args, **kwargs):
        pass
    plt.show = no_show  # disable plt show that requires the user to close the plot

    try:
        plt.rcParams.update({'figure.max_open_warning': 0})
        notebook.check_code()
        notebook.check_links()
        notebook.remove_empty_end_cell()
        notebook.check_pip_header()
        pass
    except Exception as e:
        raise Exception(notebook.filename + " failed") from e
    finally:
        plt.close('all')
        plt.rcParams.update({'figure.max_open_warning': 20})


if __name__ == '__main__':
    print("\n".join([f.filename for f in get_notebooks()]))



# -*- coding: utf-8 -*-
"""
Created on Thu Oct 27 09:40:49 2022

@author: mikf
"""

from openmdao.core.driver import Driver, RecordingDebugging
from six import iteritems
import numpy as np
from openmdao.core.analysis_error import AnalysisError
import topfarm
import time
from openmdao.utils.concurrent import concurrent_eval
from openmdao.utils.record_util import create_local_meta


class SGDDriver(Driver):

    def __init__(self, maxiter, **kwargs):
        self.maxiter = maxiter
        super().__init__(**kwargs)

        # What we support
        self.supports['integer_design_vars'] = True
        self.supports['inequality_constraints'] = True

        # What we don't support yet
        self.supports['equality_constraints'] = False
        self.supports['multiple_objectives'] = False
        self.supports['two_sided_constraints'] = False
        self.supports['linear_constraints'] = False
        self.supports['simultaneous_derivatives'] = False
        self.supports['active_set'] = False

        self._desvar_idx = {}

    def _declare_options(self):
        """
        Declare options before kwargs are processed in the init method.
        """
        self.options.declare('max_time', 5e10,
                             desc='Maximum time in seconds (set to None to disable)')
        self.options.declare('disp', True,
                             desc='Set to False to prevent printing')
        self.options.declare('run_parallel', False,
                             desc='Flag to run in parallel or not, requires mpi')
        self.options.declare('speedupSGD')
        self.options.declare('sgd_thresh')

    def _setup_driver(self, problem):
        """
        Prepare the driver for execution.

        This is the final thing to run during setup.

        Parameters
        ----------
        problem : <Problem>
            Pointer to the containing problem.
        """
        super()._setup_driver(problem)

        comm = self._problem().comm
        if not self.options['run_parallel']:
            comm = None

        if len(self._objs) > 1:
            msg = 'SGDDriver currently does not support multiple objectives.'
            raise RuntimeError(msg)

        self.comm = comm

    def run(self):
        """
        Excute the stochastic gradient descent algorithm.

        Returns
        -------
        boolean
            Failure flag; True if failed to converge, False is successful.
        """
        model = self._problem().model

        # Size design variables.
        desvars = self._designvars
        count = 0
        for name, meta in iteritems(desvars):
            size = meta['size']
            self._desvar_idx[name] = (count, count + size)
            count += size

        lower_bound = np.empty((count, ))
        upper_bound = np.empty((count, ))

        x0 = np.empty(count)
        desvar_vals = self.get_design_var_values()

        # Figure out bounds vectors and initial design vars
        for name, meta in iteritems(desvars):
            i, j = self._desvar_idx[name]
            lower_bound[i:j] = meta['lower']
            upper_bound[i:j] = meta['upper']
            x0[i:j] = desvar_vals[name]
        maxiter = self.maxiter
        max_time = self.options['max_time'] or 1e20
        assert maxiter < 1e20 or max_time < 1e20, "maxiter or max_time must be set"

        disp = self.options['disp']

        abs2prom = model._var_abs2prom['output']
        self.abs2prom = abs2prom

        # Initial Design Vars
        desvar_vals = self.get_design_var_values()
        i = 0
        for name, meta in iteritems(self._designvars):
            size = meta['size']
            x0[i:i + size] = desvar_vals[name]
            i += size

        x1 = x0.copy()
        n_iter = 0

        desvar_info = [(abs2prom[name], *self._desvar_idx[name], lower_bound, upper_bound)
                       for name, meta in iteritems(desvars)]
        desvar_dict = {name: (x0[i:j].copy(), lbound[i:j], ubound[i:j]) for (name, i, j, lbound, ubound) in desvar_info}
        start = time.time()

        self.obj_list = list(self._objs)
        self.con_list = []
        self._dvlist = list(desvar_dict)

        for name, meta in self._cons.items():
            if meta['indices'] is not None:
                meta['size'] = size = meta['indices'].indexed_src_size
            else:
                size = meta['global_size'] if meta['distributed'] else meta['size']
            self.con_list.append(name)
        self.obj_and_con_list = self.obj_list + self.con_list

        start = time.time()
        model._solve_nonlinear()
        jac = self._compute_totals(of=self.obj_and_con_list, wrt=self._dvlist,
                                   return_format='array')
        jac_time = time.time() - start
        max_iters_from_max_time = int(max_time / jac_time)
        self.maxiter = min(self.maxiter, max_iters_from_max_time)
        j = jac[0]
        learning_rate = float(np.copy(self.learning_rate))
        self.alpha0 = np.mean(np.abs(j)) / learning_rate
        self.l0 = float(np.copy(learning_rate))
        alpha = float(np.copy(self.alpha0))

        # set delta
        def multf(t, delta):
            prod = 1
            for ii in range(t):
                prod *= 1 / (1 + delta * ii)
            return learning_rate * prod

        for _ in range(self.maxiter):
            mid = np.mean([self.lower, self.upper])
            etaM = multf(self.maxiter, mid)
            if etaM < self.gamma_min:
                self.upper = mid
            elif etaM > self.gamma_min:
                self.lower = mid
        self.mid = mid

        m = np.zeros(x0.size)
        v = np.zeros(x0.size)
        self.is_converged = False
        while (n_iter < self.maxiter) and (not self.is_converged):
            obj_value_x1, x1, alpha, learning_rate, m, v, success = self.objective_callback(x1, alpha, learning_rate, m, v, record=True)
            n_iter += 1
            if disp:
                print(n_iter, obj_value_x1)
        return False

    def objective_callback(self, x, alpha, learning_rate, m, v, record=False):
        """
        Evaluate problem objective at the requested point.

        Parameters
        ----------
        x : ndarray
            Value of design variables.
        icase : int
            Case number, used for identification when run in parallel.

        Returns
        -------
        float
            Objective value
        bool
            Success flag, True if successful
        int
            Case number, used for identification when run in parallel.
        """
        model = self._problem().model
        success = 1

        for name in self._designvars:
            i, j = self._desvar_idx[name]
            self.set_design_var(name, x[i:j])

        # Execute the model
        # if record:
        with RecordingDebugging('SGD', self.iter_count, self) as rec:
            # self.set_design_var(name, value)
            self.iter_count += 1
            try:
                model._solve_nonlinear()

            # Tell the optimizer that this is a bad point.
            except AnalysisError:
                model._clear_iprint()
                success = 0

            for name, val in iteritems(self.get_objective_values()):
                obj = val
                break

            only_cons = self.alpha0 / alpha < self.options['sgd_thresh'] and self.options['speedupSGD']

            if only_cons:
                of_list = self.con_list
            else:
                of_list = self.obj_and_con_list

            self._total_jac = None
            outputs = []
            for subsys in model._subsystems_myproc:
                # check if all requested of are accounted for
                if all(self.abs2prom[x] in outputs for x in of_list):
                    subsys.skip_linearize = True
                    # print('skipped: ' + subsys.name)
                else:
                    subsys.skip_linearize = False
                outputs += list(set([b['prom_name'] for a, b in subsys.list_outputs(val=False, prom_name=True, out_stream=None)]))

            # tic = time.time()
            jac = self._compute_totals(of=of_list, wrt=self._dvlist, return_format='array')
            # print(f'time of {of_list}: {time.time() - tic:.5f}')

            if only_cons:
                c = jac[0]
                j = np.zeros_like(c)
                if c.sum() == 0:
                    self.is_converged = True
                    return obj, x, alpha, learning_rate, m, v, success
            else:
                j = jac[0]
                c = jac[1]

            # jacobian is objective gradient + alpha C
            jacobian = j + alpha * c

            # adam
            m = self.beta1 * m + (1 - self.beta1) * jacobian
            v = self.beta2 * v + (1 - self.beta2) * jacobian ** 2
            mhat = m / (1 - self.beta1 ** self.iter_count)
            vhat = v / (1 - self.beta2 ** self.iter_count)

            # update x
            x -= learning_rate * mhat / np.sqrt(vhat)

            # update learning rate and constraint scaler
            learning_rate *= 1. / (1 + self.mid * self.iter_count)
            alpha = self.alpha0 * self.l0 / learning_rate

            # validate s
            if np.any(np.isnan(x)):
                raise Exception("NaN in design Variables")

            # Record after getting obj to assure they have
            # been gathered in MPI.
            rec.abs = 0.0
            rec.rel = 0.0

        return obj, x, alpha, learning_rate, m, v, success



import json
import os
from os.path import dirname
from os.path import join as pjoin
import re
import ssl
import sys
import matplotlib.pyplot as plt
from _io import StringIO


class Notebook():
    pip_header = """# Install TopFarm if needed
import importlib
if not importlib.util.find_spec("topfarm"):
    !pip install git+https://gitlab.windenergy.dtu.dk/TOPFARM/TopFarm2.git"""

    def __init__(self, filename):
        self.filename = filename
        try:
            self.nb = self.load_notebook(self.filename)
        except Exception as e:
            raise Exception('Error in ', os.path.relpath(filename)) from e

    def __repr__(self):
        return "hej"

    def load_notebook(self, filename):
        with open(filename, encoding='utf-8') as fid:
            nb = json.load(fid)
        return nb

    def save(self, filename=None):
        filename = filename or self.filename
        with open(filename, 'w') as fid:
            json.dump(self.nb, fid, indent=4)

    def __getitem__(self, key):
        return self.nb[key]

    def __setitem__(self, key, value):
        self.nb[key] = value

    def __getattribute__(self, name):
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            if name in self.nb.keys():
                return self.nb[name]
            raise

    def insert_markdown_cell(self, index, text):
        self.cells.insert(index, {"cell_type": "markdown",
                                  "metadata": {},
                                  "source": [x + "\n" for x in text.split("\n")]
                                  })

    def insert_code_cell(self, index, code):
        self.cells.insert(index,
                          {"cell_type": "code",
                           "execution_count": 0,
                           "metadata": {},
                              "outputs": [],
                              "source": code,
                           })

    def replace_include_tag(self):
        cells = []
        for cell in self.nb['cells']:
            if cell['cell_type'] == 'code' and len(cell['source']) > 0 and '%%include' in cell['source'][0]:
                filename = pjoin(dirname(self.filename), cell['source'][0].replace('%%include', '').strip())
                nb = Notebook(filename)
                nb.replace_include_tag()
                cells.extend(nb.cells)
            else:
                cells.append(cell)
        return cells

    def get_code(self):
        code = []
        for cell in self.cells:
            if cell['cell_type'] == "code":
                if "".join(cell['source']).strip() != "":
                    code.append("".join(cell['source']))
        return code

    def get_text(self):
        txt = []
        for cell in self.cells:
            if cell['cell_type'] == "markdown":
                if "".join(cell['source']).strip() != "":
                    txt.append("".join(cell['source']))
        return txt

    def check_code(self):
        code = "\n".join(self.get_code())

        def fix(line):
            for p in ['%', '!']:
                if line.strip().startswith(p):
                    line = line.replace(p, "pass #")
            return line

        lines = [fix(x) for x in code.split("\n")]

        # import * only allowed at module level
        # So extract and remove from code lines
        module_imports = [x for x in lines if x.startswith('from') and x.endswith('import *')]
        for x in module_imports:
            lines.remove(x)

        if len(lines) == 1 and lines[0] == '':
            return
        try:
            import contextlib

            with contextlib.redirect_stdout(StringIO()):
                with contextlib.redirect_stderr(StringIO()):
                    # execute module level imports (stored in x dict) and use a locals in execution of code
                    g, x = {}, {}
                    exec("\n".join(module_imports), g, x)

                    code_str = "def test():\n    " + "\n    ".join(lines) + "\ntest()"
                    exec(code_str, x, {})
                    plt.close('all')
        except Exception as e:
            raise type(e)("Code error in %s\n%s\n" % (self.filename, str(e))).with_traceback(sys.exc_info()[2])

    def check_links(self):
        txt = "\n".join(self.get_text())
        for link in re.finditer(r"\[([^]]*)]\(([^)]*)\)", txt):
            label, url = link.groups()
            # print(label)
            # print(url)
            if url.startswith('attachment') or '#' in url:
                continue
            if url.startswith("../_static"):
                assert os.path.isfile(os.path.join(os.path.dirname(self.filename), url))
                return

            try:
                import urllib.request
                context = ssl._create_unverified_context()
                assert urllib.request.urlopen(url, context=context).getcode() == 200
            except Exception as e:
                print("%s broken in %s\n%s" % (url, self.filename, str(e)))

                # traceback.print_exc()

        # print(txt)

    def check_pip_header(self):

        code = self.get_code()
        if not code:
            return
        if code[0].strip() != self.pip_header:
            for i, cell in enumerate(self.cells):
                if cell['cell_type'] == "code":
                    break
            self.insert_code_cell(i, self.pip_header)
            self.save()
            raise Exception("""pip install header was not present in %s.
It has now been auto insert. Please check the notebook and commit the changes""" % os.path.abspath(self.filename))

    def remove_empty_end_cell(self):
        while self.cells[-1]['cell_type'] == 'code' and all([x.strip() == "" for x in self.cells[-1]['source']]):
            self.nb['cells'] = self.cells[:-1]
            self.save()


if __name__ == '__main__':
    import topfarm
    nb = Notebook(os.path.dirname(topfarm.__file__) + '/../docs/notebooks/constraints.ipynb')
    nb.check_code()
    nb.check_links()
    nb.remove_empty_end_cell()
    nb.check_pip_header()



from openmdao.core.explicitcomponent import ExplicitComponent
import numpy as np
import time
from _collections import defaultdict
from topfarm.constraint_components.post_constraint import PostConstraint
import warnings


class CostModelComponent(ExplicitComponent):
    """Wrapper for pure-Python cost functions"""

    def __init__(self, input_keys, n_wt, cost_function, cost_gradient_function=None,
                 output_keys=["Cost"], output_unit="", additional_input=[], additional_output=[], max_eval=None,
                 objective=True, maximize=False, output_vals=[0.0], input_units=[], step={}, use_penalty=True, **kwargs):
        """Initialize wrapper for pure-Python cost function

        Parameters
        ----------
        input_keys : list of str
            Inputs to the cost function
        n_wt : int
            Number of wind turbines
        cost_function : function handle
            Function to evaluate cost
        cost_gradient_function : function handle, optional
            Function to evaluate gradient of the cost function
        output_key : list
            Name of output key
        output_unit : str
            Units of output of cost function
        additional_input : list of str
            Other (non-design-variable) inputs required by the cost function\n
            Gradients will not be computed for these inputs
        additional_output : list of str or list of tuples
            Other outputs to request\n
            if list of str: ['add_out_name',...]\n
            if list of tuples [('add_out_name', val),...], where val is a template value of the output
            The cost function must return: cost, {'add_out1_name': add_out1, ...}
        max_eval : int
            Maximum number of function evaluations
        objective : boolean
            Must be True for standalone CostModelComponents and the final component in a TopFarmGroup
        maximize : boolean
            If True: objective is maximised during optimization\n
            If False: Objective is minimized during optimization
        output_vals : float or array_like
            Format of output
        input_units : list of str
            Units of the respective input_keys
        step : dict of {str : float}
            Finite difference step size for each input key, e.g. {input_key : step_size, input_key2 : step_size2}
        """
        if 'income_model' in kwargs:
            warnings.warn("""income_model is deprecated; use keyword maximize instead""",
                          DeprecationWarning, stacklevel=2)
            maximize = kwargs['income_model']
            kwargs.pop('income_model')
        if 'output_key' in kwargs:
            warnings.warn("""output_key is deprecated; use keyword output_keys instead""",
                          DeprecationWarning, stacklevel=2)
            output_keys = [kwargs['output_key']]
            kwargs.pop('output_key')
        else:
            self.output_key = output_keys[0]
        if 'output_val' in kwargs:
            warnings.warn("""output_val is deprecated; use keyword output_vals instead""",
                          DeprecationWarning, stacklevel=2)
            output_vals = [kwargs['output_val']]
            kwargs.pop('output_val')
        super().__init__(**kwargs)
        assert isinstance(n_wt, int), n_wt
        self.input_keys = list(input_keys)
        self.input_keys_only = list([(i, i[0])[isinstance(i, tuple)] for i in self.input_keys])
        self.additional_input = additional_input
        self.all_input_keys = self.input_keys_only + list([(i, i[0])[isinstance(i, tuple)] for i in self.additional_input])
        self.cost_function = cost_function
        self.cost_gradient_function = cost_gradient_function
        self.n_wt = n_wt
        if not isinstance(output_keys, list):
            output_keys = [output_keys]
        if not isinstance(output_vals, list):
            output_vals = [output_vals]
        self.output_keys = output_keys
        self.out_keys_only = list([(o, o[0])[isinstance(o, tuple)] for o in self.output_keys])
        self.output_unit = output_unit
        self.output_key = output_keys[0]
        if isinstance(self.output_key, tuple):
            self.output_key = self.output_key[0]
        self.additional_output = [((x, np.zeros(self.n_wt)), x)[isinstance(x, tuple)] for x in additional_output]
        self.max_eval = max_eval or 1e100
        self.objective = objective
        if maximize:
            self.cost_factor = -1.0
        else:
            self.cost_factor = 1.0
        output_vals = output_vals[:len(output_keys)] + [output_vals[0]] * (len(output_keys) - len(output_vals))  # extend output_vals if it is not same length as output_keys
        self.output_vals = output_vals

        n_input = len(self.input_keys) + len(additional_input)
        self.input_units = (input_units + [None] * n_input)[:n_input]

        self.n_func_eval = 0
        self.func_time_sum = 0
        self.n_grad_eval = 0
        self.grad_time_sum = 0
        self.step = step
        self.use_penalty = use_penalty

    def setup(self):
        for i, u in zip(self.input_keys + self.additional_input, self.input_units):
            if isinstance(i, tuple) and len(i) == 2:
                self.add_input(i[0], val=i[1], units=u)
            else:
                self.add_input(i, val=np.zeros(self.n_wt), units=u)
        if self.use_penalty:
            self.add_input('penalty', val=0.0)
        if self.objective:
            self.add_output('cost', val=0.0)
            self.add_output('cost_comp_eval', val=0.0)
        for o, v in zip(self.output_keys, self.output_vals):
            if isinstance(o, tuple) and len(o) == 2:
                self.add_output(o[0], val=o[1])
            else:
                self.add_output(o, val=v)
        for key, val in self.additional_output:
            self.add_output(key, val=val)

        if self.cost_gradient_function:
            if self.objective:
                self.declare_partials('cost', self.input_keys_only, method='exact')
            for o in self.out_keys_only:
                self.declare_partials(o, self.input_keys_only, method='exact')
        else:
            if self.step == {}:
                if self.objective:
                    self.declare_partials('cost', self.input_keys_only, method='fd')
                for o in self.out_keys_only:
                    self.declare_partials(o, self.input_keys_only, method='fd')
            else:
                for i in self.input_keys_only:
                    if self.objective:
                        self.declare_partials('cost', i, step=self.step[i], method='fd')
                    for o in self.out_keys_only:
                        self.declare_partials(o, i, step=self.step[i], method='fd')

    @property
    def counter(self):
        counter = float(self.n_func_eval)
        if self.grad_time_sum > 0 and self.func_time_sum > 0 and self.n_grad_eval > 0 and self.n_func_eval > 0:
            ratio = ((self.grad_time_sum / self.n_grad_eval) /
                     (self.func_time_sum / self.n_func_eval))
            counter += self.n_grad_eval * max(ratio, 1)
        else:
            counter += self.n_grad_eval
        return int(counter)

    def compute(self, inputs, outputs):
        """Compute cost model"""
        if self.use_penalty:
            if inputs['penalty'] > 0:
                return
        if self.counter >= self.max_eval:
            return
        t = time.time()
        if self.additional_output:
            c, additional_output = self.cost_function(**{x: inputs[x] for x in self.all_input_keys})
            for k, v in additional_output.items():
                outputs[k] = v
        else:
            c = self.cost_function(**{x: inputs[x] for x in self.all_input_keys})
        if not isinstance(c, list):
            c = [c]
        if self.objective:
            outputs['cost'] = c[0] * self.cost_factor
            outputs['cost_comp_eval'] = self.counter
        for o, _c in zip(self.out_keys_only, c):
            outputs[o] = _c
        self.func_time_sum += time.time() - t
        self.n_func_eval += 1

    def compute_partials(self, inputs, J):
        if self.counter >= self.max_eval:
            return

        if hasattr(self, 'skip_linearize'):
            if self.skip_linearize:
                return

        t = time.time()
        if self.cost_gradient_function:
            for k, dCostdk in zip(self.input_keys_only,
                                  self.cost_gradient_function(**{x: inputs[x] for x in self.all_input_keys})):
                if dCostdk is not None:
                    if not isinstance(dCostdk, list):
                        dCostdk = [dCostdk]
                    if self.objective:
                        J['cost', k] = dCostdk[0] * self.cost_factor
                    for o, _d in zip(self.out_keys_only, dCostdk):
                        J[o, k] = _d
        self.grad_time_sum += time.time() - t
        self.n_grad_eval += 1


class AEPCostModelComponent(CostModelComponent):
    """Wrapper for pure-Python cost functions"""
    def __init__(self, input_keys, n_wt, cost_function, cost_gradient_function=None,
                 output_unit="", additional_input=[], additional_output=[], max_eval=None,
                 output_key="AEP", **kwargs):
        """Initialize Sub class of the CostModelComponent for AEP maximization

        Parameters
        ----------
        input_keys : list of str
            Inputs to the cost function
        n_wt : int
            Number of wind turbines
        cost_function : function handle
            Function to evaluate cost
        cost_gradient_function : function handle, optional
            Function to evaluate gradient of the cost function
        output_key : list
            Name of output key, default is AEP
        output_unit : str
            Units of output of cost function
        additional_input : list of str
            Other (non-design-variable) inputs required by the cost function\n
            Gradients will not be computed for these inputs
        additional_output : list of str or list of tuples
            Other outputs to request\n
            if list of str: ['add_out_name',...]\n
            if list of tuples [('add_out_name', val),...], where val is a template value of the output
            The cost function must return: cost, {'add_out1_name': add_out1, ...}
        max_eval : int
            Maximum number of function evaluations
        """
        CostModelComponent.__init__(self, input_keys, n_wt, cost_function,
                                    cost_gradient_function=cost_gradient_function,
                                    output_key=output_key, output_unit=output_unit,
                                    additional_input=additional_input,
                                    additional_output=additional_output,
                                    max_eval=max_eval, maximize=True, **kwargs)


class AEPMaxLoadCostModelComponent(CostModelComponent, PostConstraint):
    """Wrapper for pure-Python cost functions"""
    def __init__(self, input_keys, n_wt, aep_load_function, max_loads,
                 aep_load_gradient=None, output_keys=["AEP", 'loads'], step={},
                 maximize=True, **kwargs):
        """Initialize Sub class of the CostModelComponent for AEP maximization and load constraints

        Parameters
        ----------
        input_keys : list of str
            Inputs to the cost function
        n_wt : int
            Number of wind turbines
        aep_load_function : function handle
            Function to evaluate cost
        aep_load_gradient : function handle, optional
            Function to evaluate gradient of the cost function
        output_keys : list
            Name of output key, default is AEP and loads
        step : dict of {str : float}
            Finite difference step size for each input key, e.g. {input_key : step_size, input_key2 : step_size2}
        maximize : boolean
            If True: objective is maximised during optimization\n
            If False: Objective is minimized during optimization
        """

        self.max_loads = max_loads

        def cost_function(**kwargs):
            aep, load = aep_load_function(**kwargs)
            return [aep, load]

        if aep_load_gradient:
            def gradient_function(**kwargs):
                d_aep, d_load = aep_load_gradient(**kwargs)
                return [[d_aep, d_load]]
        else:
            gradient_function = None

        # additional_output = kwargs.get('additional_output', []) + [('loads', max_loads)]
        CostModelComponent.__init__(self, input_keys, n_wt, cost_function=cost_function,
                                    cost_gradient_function=gradient_function,
                                    output_keys=output_keys, step=step, maximize=maximize,
                                    **kwargs)
        PostConstraint.__init__(self, 'loads', upper=max_loads)

    # def setup(self):
    #     AEPCostModelComponent.setup(self)
    #     input_keys = list([(i, i[0])[isinstance(i, tuple)] for i in self.input_keys])
        # self.declare_partials('loads', input_keys, method='fd')
